/*
** cmap.c -- Colormap routines
**
** These routines are architecture dependant.
** Make sure the appropriate architecture macros are #defined
**
** Copyright (C) 1992, 1993, 1994 by Michael J. Gourlay
**
** NO WARRANTEES, EXPRESS OR IMPLIED.
*/

#include "showtime.h"
#include "cmap.h"
#include "video.h"
#include "pbm.h"

#include "hsv.c"

/* ---------------------------------------------------------------------- */
/* Global variables */

/* for NTSC -- saturation scalar */
float gm_ntsc=1.0;

/* force CGA palette (off by default) */
int gm_force_cga=0;

#ifdef X11
#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

/* X11 Colormap variables */
int           c_cmap=0; /* whether there was a colormap allocated */
Colormap      x_cmap;
unsigned long pixels[256];
int           pal1to1=1; /* flag: whether cmap index is array index */
#endif /* X11 */

/* ---------------------------------------------------------------------- */
/*
** IBM PC architecture data
**
** This module contains data to simulate, to some degree, the
** video behaviour of obsolete IBM PC video cards.
*/

/* cga_color is a list of RGB values for the CGA colors */
unsigned char cga_color[][3] = {
  {0,0,0},    {0,0,127},   {0,127,0},   {0,127,127},
  {127,0,0},  {127,0,127}, {127,127,0}, {127,127,127},
  {63,63,63}, {0,0,255},   {0,255,0},   {0,255,255},
  {255,0,0},  {255,0,255}, {255,255,0}, {255,255,255},
};

/* ---------------------------------------------------------------------- */

/* cga_palette is a list of CGA colors for the lame CGA palettes */
/* Each CGA palette consists of a border color plus 3 CGA colors */
int cga_palette[][3] = {
    {2,4,6}, {3,5,15}, {3,4,15}, {10,12,14}, {11,13,15}, {11,12,15}
};

/* ====================================================================== */

#define RM (7<<5)
#define GM (7<<2)
#define BM (3)

/* make_palette : create a palette
** 25mar94 MJG
**
** approx tells whether to convert to B/W or 332 RGB cmap
*/
int
make_palette(int approx, unsigned char rgb_map[][3])
{
  register int pe;

  for(pe=0; pe<256; pe++) {
    if(approx==GRAY) {
      rgb_map[pe][R]=rgb_map[pe][G]=rgb_map[pe][B] = pe;
    } else if(approx==RGB332) {
      register int ri, gi, bi;
      /* 3 bits red, 3 bits green, 2 bits blue */
      ri = (((pe & RM)>>5) * 255 + 3) / 7;
      gi = (((pe & GM)>>2) * 255 + 3) / 7;
      bi = ( (pe & BM)     * 255 + 1) / 3;
      rgb_map[pe][R] = ri;
      rgb_map[pe][G] = gi;
      rgb_map[pe][B] = bi;
    } else {
      logmesg("make_palette: invalid approximation method\n");
    }
  }
  return(0);
}

/* ---------------------------------------------------------------------- */

#define CGA_COLORS 16

/* My copy of the colormap in RGB order */
static unsigned char st_cmap[256][3];

/* MicroScum's manual is WRONG about how to load the palette */
#if defined(MSC)
/*#define MAP_RGB(ri, gi, bi) (((long) ((bi)<<8 | (gi))<<8) | (ri))/* WRONG */
#define MAP_RGB(ri, gi, bi) ( ((long)(bi)<<14) | ((long)(gi)<<6) | ((ri)>>2) )
#endif

/* install_cmap : install a colormap
** numvals is the number of R,G,B triples
** rgb_map is an array of R,G,B values, in the range 0-maxval
*/
int
install_cmap(int num_vals, unsigned char rgb_map[][3], int maxval)
{
  register int pe;
  unsigned char red, green, blue;
  float    rf, gf, bf; /* floating point versions for HSV conversion */
  float    hf, sf, vf; /* HSV color space values */

#if defined(MSC)
  static long col[256];
#elif defined(X11)
  XColor color;
#endif /* architecture */

#if defined(X11)

  if(x_depth==8) {
    x_cmap=XCreateColormap(x_display, x_window,
        DefaultVisual(x_display, x_screen), AllocNone);
    c_cmap=1; /* set flag: X colormap was allocated */
  } else {
    logmesg("depth %i not supported\n", x_depth);
    x_cmap=DefaultColormap(x_display, x_screen);
  }

#endif

  /* Read each palette color value and install it */
  for(pe=0; pe<256; pe++) {

    /* Install the new colors */
    if(pe<num_vals) {

      /* Force CGA colors for first 16 palette entries */
      if(gm_force_cga && pe<CGA_COLORS) {
        red=cga_color[pe][0];
        green=cga_color[pe][1];
        blue=cga_color[pe][2];
        rf = (float)red / 255;
        gf = (float)green / 255;
        bf = (float)blue / 255;
      } else {
        /* Scale the color values to be from 0 to 255 */
        rf = (float)rgb_map[pe][0] / maxval;
        gf = (float)rgb_map[pe][1] / maxval;
        bf = (float)rgb_map[pe][2] / maxval;
        red   = rf * 255 + 0.5;
        green = gf * 255 + 0.5;
        blue  = bf * 255 + 0.5;
      }

    /* At this step, the working color is rf, gf, bf */

      /* Scale the saturation */
      if(gm_ntsc!=1.0) {
        rgb_to_hsv(rf, gf, bf, &hf, &sf, &vf);
        sf *= gm_ntsc;
        hsv_to_rgb(&rf, &gf, &bf, hf, sf, vf);
        red   = rf * 255 + 0.5;
        green = gf * 255 + 0.5;
        blue  = bf * 255 + 0.5;
      }

    /* At this step, the working color is st_cmap[pe][R|G|B] */

      /* Store the colormap in our private map */
      st_cmap[pe][R] = red;
      st_cmap[pe][G] = green;
      st_cmap[pe][B] = blue;
    } else {
      /* Load colors from private map */
      red   = st_cmap[pe][R];
      green = st_cmap[pe][G];
      blue  = st_cmap[pe][B];
    }


    /* Rescale the palette values for the device */

#if defined(MSC)
    /* Convert color value into a VGA palette 32 bit color value */
    col[pe] = MAP_RGB(red, green, blue);

    /* Map the corresponding colormap index to the color value */
    /*_remappalette(pe, col[pe]);/**/

#elif defined(TC)
    setrgbpalette(pe, red, green, blue);

#elif defined(X11)
    color.red   = red << 8;
    color.green = green << 8;
    color.blue  = blue << 8;
    if(!XAllocColor(x_display, x_cmap, &color)) {
      logmesg("did not allocate color %i:%i,%i,%i\n", pe, red, green, blue);
    }
    pixels[pe]=color.pixel;

    /* See if the palette lookup table value is its index */
    /* This is for a speedup optimization in the image display */
    if(pal1to1 && pixels[pe]!=pe) pal1to1=0;

#endif /* architecture */
  }

  /* Install the colormap internally */

#if defined(MSC)
  _remapallpalette(col); /**/

#elif defined(X11)

  if(x_depth==8) {
    XSetWindowColormap(x_display, x_window, x_cmap);
  }
#ifdef XSYNC_CMAP
  XSync(x_display, /* discard events */False);
#endif

#endif /* architecture */

  return(0);
}
